@@ -22,7 +22,7 @@
 #include <stdio.h>
 #include <string.h>
 #include <errno.h>
-#include <math.h>
+#include <limits.h>
 #if ENABLE_NLS
 # include "gettext.h"
 #else
@@ -38,6 +38,8 @@
 #endif
 #include "json.h"
 
+enum json_num_type json_num_type = JSON_NUM_DOUBLE;
+
 void *
 json_2nrealloc(void *p, size_t *pn, size_t s)
 {
@@ -95,8 +97,12 @@ json_strerror(int ec)
 		[JSON_E_NOMEM]     = N_("Not enough memory"),
 		[JSON_E_BADTOK]    = N_("Unrecognized token"),
 		[JSON_E_BADDELIM]  = N_("Bad delimiter"),
-		[JSON_E_BADSTRING] = N_("Malformed string")
+		[JSON_E_BADSTRING] = N_("Malformed string"),
+		[JSON_E_RANGE]     = N_("Value out of range"),
+		[JSON_E_EOF]       = N_("Unexpected end of input"),
+		[JSON_E_USAGE]     = N_("Usage error")
 	};
+	ec &= ~JSON_E_MOREINPUT;
 	if (ec < 0 || ec >= sizeof(json_errstr) / sizeof(json_errstr[0]))
 		return gettext("Unknown error");
 	return gettext(json_errstr[ec]);
@@ -162,8 +168,10 @@ static void json_format_null(struct json_format *, struct json_value *,
 			     size_t level);
 static void json_format_bool(struct json_format *, struct json_value *,
 			     size_t level);
-static void json_format_number(struct json_format *, struct json_value *,
+static void json_format_double(struct json_format *, struct json_value *,
 			       size_t level);
+static void json_format_long(struct json_format *, struct json_value *,
+			     size_t level);
 static void json_format_string(struct json_format *, struct json_value *,
 			       size_t level);
 static void json_format_array(struct json_format *, struct json_value *,
@@ -183,7 +191,8 @@ static struct json_value_meth {
 } json_value_meth[] = {
 	[json_null]   = { json_format_null, NULL, json_copy_generic },
 	[json_bool]   = { json_format_bool, NULL, json_copy_generic },
-	[json_number] = { json_format_number, NULL, json_copy_generic },
+	[json_double] = { json_format_double, NULL, json_copy_generic },
+	[json_long]   = { json_format_long, NULL, json_copy_generic },
 	[json_string] = { json_format_string, json_free_string, json_copy_string },
 	[json_array]  = { json_format_array, json_free_array, json_copy_array },
 	[json_object] = { json_format_object, json_free_object, json_copy_object }
@@ -292,18 +301,18 @@ json_format_bool(struct json_format *fmt, struct json_value *val, size_t level)
 	json_format_writez(fmt, val->b ? "true" : "false");
 }
 
-/* Number value */
+/* Double-precision numeric value */
 struct json_value *
-json_new_number(double n)
+json_new_double(double n)
 {
-	struct json_value *j = json_value_create(json_number);
+	struct json_value *j = json_value_create(json_double);
 	if (j)
 		j->n = n;
 	return j;
 }
 
 static void
-json_format_number(struct json_format *fmt, struct json_value *val, size_t level)
+json_format_double(struct json_format *fmt, struct json_value *val, size_t level)
 {
 	char buffer[128];//FIXME
 	if (fmt->precision == -1)
@@ -314,6 +323,24 @@ json_format_number(struct json_format *fmt, struct json_value *val, size_t level
 	json_format_writez(fmt, buffer);
 }
 
+/* Long integer value */
+struct json_value *
+json_new_long(long n)
+{
+	struct json_value *j = json_value_create(json_long);
+	if (j)
+		j->i = n;
+	return j;
+}
+
+static void
+json_format_long(struct json_format *fmt, struct json_value *val, size_t level)
+{
+	char buffer[128];//FIXME
+	snprintf(buffer, sizeof buffer, "%ld", val->i);
+	json_format_writez(fmt, buffer);
+}
+
 /* String value */
 struct json_value *
 json_new_string(char const *str)
@@ -786,16 +813,37 @@ json_object_filter(struct json_value *obj,
 #define ISHEX(c) (strchr("0123456789abcdefABCDEF", c) != NULL)
 #define SKIPWS(s, e) while ((s) != (e) && ISSPACE(*(s))) (s)++;
 
+enum state {
+	ST_INIT,
+	ST_KEY,    /* Object: key read. */
+	ST_COLON,  /* Object: key and colon obtained. */
+	ST_VALUE,  /* Object, Array: value read. */
+	ST_COMMA,  /* Object, Array: comma read. */
+};
+
 struct j_context {
 	struct j_context *next;
 	struct json_value *obj;
 	char *key;
+	enum state state;
 };
 
 static inline int
-j_context_type(struct j_context *ctx)
+j_context_type(struct j_context **ctx)
 {
-	return ctx ? ctx->obj->type : json_null;
+	return *ctx ? (*ctx)->obj->type : json_null;
+}
+
+static inline enum state
+j_context_state(struct j_context **ctx)
+{
+	return *ctx ? (*ctx)->state : ST_INIT;
+}
+
+static inline void
+j_context_set_state(struct j_context **ctx, enum state state)
+{
+	(*ctx)->state = state;
 }
 
 static int
@@ -818,8 +866,9 @@ j_context_push(struct j_context **ctx, int type)
 		free(cur);
 		return JSON_E_NOMEM;
 	}
-	cur->next = *ctx;
 	cur->key = NULL;
+	cur->state = ST_INIT;
+	cur->next = *ctx;
 	*ctx = cur;
 	return JSON_E_NOERR;
 }
@@ -833,7 +882,8 @@ j_context_pop(struct j_context **ctx)
 	if (!*ctx)
 		return NULL;
 	val = (*ctx)->obj;
-	free((*ctx)->key);
+	if (val->type == json_object)
+		free((*ctx)->key);
 	next = (*ctx)->next;
 	free(*ctx);
 	*ctx = next;
@@ -915,6 +965,16 @@ utf8_wctomb(char const *u, char r[6], char **endp)
 	return count;
 }
 
+static int
+memspn(char const *p, char const *delim, char const *stop)
+{
+	int i;
+	for (i = 0; p != stop; i++, p++)
+		if (!strchr(delim, *p))
+			break;
+	return i;
+}
+
 static int
 memcspn(char const *p, char const *delim, char const *stop)
 {
@@ -932,15 +992,29 @@ j_get_text(char const *input, char const *stop, char **retval,
 	char *str = NULL;
 	size_t len = 0;
 	size_t cap = 0;
-	int ec = JSON_E_NOERR;
+	int ec = JSON_E_BADSTRING|JSON_E_MOREINPUT;
 	int tl = 0; /* Length of the tail segment of input. */
 
 	input++;
-	while (*input != '"') {
-		int n = memcspn(input, "\\\"", stop);
+	while (input < stop) {
+		int n;
+
+		if (*input == '"') {
+			ec = JSON_E_NOERR;
+			break;
+		}
+
+		n = memcspn(input, "\\\"", stop);
+
+		if (input + n == stop) {
+			ec = JSON_E_BADSTRING|JSON_E_MOREINPUT;
+			break;
+		}
+
 		/* Optimize memory allocations for the common case: */
 		if (input[n] == '"') {
 			tl = n;
+			ec = JSON_E_NOERR;
 			break;
 		}
 		if (strexpand(&str, &cap, len, n)) {
@@ -951,26 +1025,29 @@ j_get_text(char const *input, char const *stop, char **retval,
 		input += n;
 		len += n;
 
-		if (input == stop) {
-			ec = JSON_E_BADSTRING;
-			break;
-		}
-
 		if (*input == '\\') {
 			char r[6];
-			input++;
+			if (++input == stop) {
+				ec = JSON_E_BADSTRING|JSON_E_MOREINPUT;
+				break;
+			}
 			if ((r[0] = escape(*input, UNESCAPE)) != 0) {
 				n = 1;
 				input++;
-			} else if (*input == 'u' && input + 5 < stop) {
-				char *errp;
-				n = utf8_wctomb(input + 1, r, &errp);
-				if (n == -1) {
-					ec = JSON_E_BADSTRING;
-					input = errp;
+			} else if (*input == 'u') {
+				if (input + 5 < stop) {
+					char *errp;
+					n = utf8_wctomb(input + 1, r, &errp);
+					if (n == -1) {
+						ec = JSON_E_BADSTRING;
+						input = errp;
+						break;
+					}
+					input += 5;
+				} else {
+					ec = JSON_E_BADSTRING|JSON_E_MOREINPUT;
 					break;
 				}
-				input += 5;
 			} else {
 				ec = JSON_E_BADSTRING;
 				break;
@@ -1002,123 +1079,292 @@ j_get_text(char const *input, char const *stop, char **retval,
 	return ec;
 }
 
+static int
+mem_to_double(char const *ptr, char const *stop,
+	      struct json_value **valp, char **endp)
+{
+	size_t n = 0;
+	char *store, *p;
+	double d;
+
+	switch (ptr[n]) {
+	case '-':
+	case '+':
+		n++;
+	}
+
+	n += memspn(ptr+n, ".0123456789", stop);
+	if (ptr[n] == 'e' || ptr[n] == 'E') {
+		n++;
+		if (ptr[n] == '-')
+			n++;
+		n += memspn(ptr+n, "0123456789", stop);
+	}
+
+	store = strndup(ptr, n);
+	if (!store)
+		return JSON_E_NOMEM;
+	errno = 0;
+	d = strtod(store, &p);
+	n = p - store;
+	free(store);
+	*endp = (char*)ptr + n;
+	if (errno == ERANGE)
+		return JSON_E_RANGE;
+	if ((*valp = json_new_double(d)) == NULL)
+		return JSON_E_NOMEM;
+	return JSON_E_NOERR;
+}
+
+static int
+mem_to_long(char const *ptr, char const *stop,
+	    struct json_value **valp, char **endp)
+{
+	int neg;
+	long n = 0;
+
+	switch (*ptr) {
+	case '-':
+		neg = 1;
+		ptr++;
+		break;
+	case '+':
+		neg = 0;
+		ptr++;
+		break;
+	default:
+		neg = 0;
+	}
+
+	for (;ptr < stop; ptr++) {
+		int d = *ptr - '0';
+		if (d < 0 || d > 9) {
+			*endp = (char*)ptr;
+			break;
+		}
+		if (neg ? (LONG_MIN / 10 > -n) : (LONG_MAX / 10 < n)) {
+			*endp = (char*)ptr;
+			return JSON_E_RANGE;
+		}
+		n *= 10;
+		if (neg ? (LONG_MIN + d > -n) : (LONG_MAX - d < n)) {
+			*endp = (char*)ptr;
+			return JSON_E_RANGE;
+		}
+		n += d;
+	}
+
+	if (neg)
+		n = - n;
+
+	*endp = (char*)ptr;
+
+	if ((*valp = json_new_long(n)) == NULL)
+		return JSON_E_NOMEM;
+
+	return JSON_E_NOERR;
+}
+
+static int
+memton(char const *ptr, char const *stop,
+       struct json_value **valp, char **endp)
+{
+	int rc;
+	struct json_value *v;
+	char *ep;
+
+	switch (json_num_type) {
+	case JSON_NUM_DOUBLE:
+		rc = mem_to_double(ptr, stop, valp, endp);
+		break;
+	case JSON_NUM_LONG:
+	case JSON_NUM_AUTO:
+		if ((rc = mem_to_long(ptr, stop, &v, &ep)) == JSON_E_NOERR) {
+			if (strchr(".eE", *ep)) {
+				json_value_free(v);
+				rc = mem_to_double(ptr, stop, &v, &ep);
+			}
+		} else if (rc == JSON_E_RANGE)
+			rc = mem_to_double(ptr, stop, &v, &ep);
+		if (rc == JSON_E_NOERR && json_is_double(v)) {
+			long i;
+			if (LONG_MIN <= v->n && v->n <= LONG_MAX &&
+			    (i = v->n) == v->n) {
+				/* Convert to long */
+				v->i = i;
+				v->type = json_long;
+				*valp = v;
+				*endp = ep;
+			} else if (json_num_type == JSON_NUM_LONG) {
+				json_value_free(v);
+				rc = JSON_E_RANGE;
+				*endp = (char*)ptr;
+			} else {
+				*valp = v;
+				*endp = ep;
+			}
+		} else {
+			*valp = v;
+			*endp = ep;
+		}
+		break;
+	}
+	return rc;
+}
+
 int
 json__parse(struct j_context **pctx, char const *input, char const *stop,
-	   struct json_value **retval, char **endp)
+	    struct json_value **retval, char **endp, char **tok)
 {
 	struct json_value *val;
 	int ecode;
 	char *str;
-# define RETERR(c) do { *endp = (char*)input; return c; } while(0)
-
+	char const *p;
+# define RETERR(c) do { if (endp) *endp = (char*)input; return c; } while(0)
+# define SAVETOK() do { if (tok) *tok = (char*)input; } while (0)
 	while (1) {
 		val = NULL;
 
 		SKIPWS(input, stop);
-		if (input == stop)
-			RETERR(JSON_E_BADDELIM);
+		SAVETOK();
+		if (input == stop) {
+			RETERR(JSON_E_EOF|JSON_E_MOREINPUT);
+		}
 
-		switch (*input) {
-		case '[':
-			j_context_push(pctx, json_array);
-			++input;
-			continue;
-
-		case ']':
-			if (j_context_type(*pctx) == json_array)
-				val = j_context_pop(pctx);
-			else
-				RETERR(JSON_E_BADTOK);
-			++input;
-			break;
-
-		case '{':
-			j_context_push(pctx, json_object);
-			++input;
-			continue;
-
-		case '}':
-			if (j_context_type(*pctx) == json_object)
-				val = j_context_pop(pctx);
-			else
-				RETERR(JSON_E_BADTOK);
-			++input;
-			break;
-
-		case '"':
-			ecode = j_get_text(input, stop, &str, &input);
-			if (ecode != JSON_E_NOERR)
-				RETERR(ecode);
-			input++;
-
-			if (j_context_type(*pctx) == json_object &&
-			    !(*pctx)->key) {
-				(*pctx)->key = str;
-
-				SKIPWS(input, stop);
-				if (input != stop && *input == ':') {
+		switch (j_context_state(pctx)) {
+		case ST_VALUE:
+			if (*input == ',') {
+				j_context_set_state(pctx, ST_COMMA);
+				input++;
+				continue;
+			}
+			/* fall through */
+		case ST_INIT:
+			switch (*input) {
+			case ']':
+				if (j_context_type(pctx) == json_array) {
+					val = j_context_pop(pctx);
+					if (*pctx)
+						j_context_set_state(pctx,
+								    ST_VALUE);
 					++input;
+					goto getval;
+				} else
+					RETERR(JSON_E_BADTOK);
+				break;
+
+			case '}':
+				if (j_context_type(pctx) == json_object) {
+					val = j_context_pop(pctx);
+					if (*pctx)
+						j_context_set_state(pctx,
+								    ST_VALUE);
+					++input;
+					goto getval;
+				} else
+					RETERR(JSON_E_BADTOK);
+				break;
+			}
+			/* fall through */
+
+		case ST_COLON:
+		case ST_COMMA:
+			switch (*input) {
+			case '[':
+				j_context_push(pctx, json_array);
+				++input;
+				continue;
+
+			case '{':
+				j_context_push(pctx, json_object);
+				++input;
+				continue;
+
+			case '"':
+				SAVETOK();
+				ecode = j_get_text(input, stop, &str, &p);
+				if (ecode != JSON_E_NOERR)
+					RETERR(ecode);
+				input = p + 1;
+				if (j_context_type(pctx) == json_object &&
+				    (j_context_state(pctx) == ST_INIT ||
+				     j_context_state(pctx) == ST_COMMA)) {
+					(*pctx)->key = str;
+					j_context_set_state(pctx, ST_KEY);
 					continue;
+				} else {
+					val = json_value_create(json_string);
+					if (!val)
+						RETERR(JSON_E_NOMEM);
+					val->s = str;
 				}
-				RETERR(JSON_E_BADDELIM);
-			} else {
-				val = json_value_create(json_string);
+				break;
+
+			case '-':
+			case '+':
+			case '.':
+			case '0':
+			case '1':
+			case '2':
+			case '3':
+			case '4':
+			case '5':
+			case '6':
+			case '7':
+			case '8':
+			case '9':{
+				char *p;
+
+				SAVETOK();
+				ecode = memton(input, stop, &val, &p);
+				if (ecode != JSON_E_NOERR)
+					RETERR(ecode);
+				input = p;
+				if (p == stop) {
+					if (*pctx) {
+						json_value_free(val);
+						RETERR(JSON_E_EOF|JSON_E_MOREINPUT);
+					}
+				}
+				break;
+			}
+
+			default:{
+				size_t n = stop - input;
+				if (n >= 4 && memcmp(input, "null", 4) == 0) {
+					val = json_new_null();
+					input += 4;
+				} else if (n >= 4 && memcmp(input, "true", 4) == 0) {
+					val = json_new_bool(1);
+					input += 4;
+				} else if (n >= 5 && memcmp(input, "false", 5) == 0) {
+					val = json_new_bool(0);
+					input += 5;
+				} else
+					RETERR(JSON_E_BADTOK|JSON_E_MOREINPUT);//FIXME
 				if (!val)
 					RETERR(JSON_E_NOMEM);
-				val->s = str;
+			  }
 			}
 			break;
 
-		case '-':
-		case '+':
-		case '.':
-		case '0':
-		case '1':
-		case '2':
-		case '3':
-		case '4':
-		case '5':
-		case '6':
-		case '7':
-		case '8':
-		case '9': {
-			char *p;
-			double d = strtod(input, &p);
-			if (p > stop) //FIXME
-				RETERR(JSON_E_BADTOK);
-			if ((input[0] == '-' && d == -HUGE_VAL)
-			    || d == HUGE_VAL)
-				RETERR(JSON_E_BADTOK);
-			val = json_new_number(d);
-			if (!val)
-				RETERR(JSON_E_NOMEM);
-			input = p;
-			break;
-		}
+		case ST_KEY:
+			if (*input == ':') {
+				j_context_set_state(pctx, ST_COLON);
+				input++;
+				continue;
+			}
+			RETERR(JSON_E_BADDELIM);
 
-		default:{
-			size_t n = stop - input;
-			if (n >= 4 && memcmp(input, "null", 4) == 0) {
-				val = json_new_null();
-				input += 4;
-			} else if (n >= 4 && memcmp(input, "true", 4) == 0) {
-				val = json_new_bool(1);
-				input += 4;
-			} else if (n >= 5 && memcmp(input, "false", 5) == 0) {
-				val = json_new_bool(0);
-				input += 5;
-			} else
-				RETERR(JSON_E_BADTOK);
-			if (!val)
-				RETERR(JSON_E_NOMEM);
-		  }
+		default:
+			abort();
 		}
-
+	getval:
 		if (val) {
 			if (*pctx) {
 				int rc;
 
-				switch (j_context_type(*pctx)) {
+				switch (j_context_type(pctx)) {
 				case json_array:
 					rc = json_array_append((*pctx)->obj,
 							       val);
@@ -1127,7 +1373,7 @@ json__parse(struct j_context **pctx, char const *input, char const *stop,
 				case json_object:
 					if (!(*pctx)->key) {
 						json_value_free(val);
-						return JSON_E_BADTOK;
+						RETERR(JSON_E_BADTOK);
 					}
 					rc = json_object_set((*pctx)->obj,
 							     (*pctx)->key,
@@ -1143,40 +1389,70 @@ json__parse(struct j_context **pctx, char const *input, char const *stop,
 					json_value_free(val);
 					RETERR(JSON_E_NOMEM);
 				}
+				j_context_set_state(pctx, ST_VALUE);
 			} else {
+				SKIPWS(input, stop);
+				if (input < stop) {
+					json_value_free(val);
+					RETERR(JSON_E_BADTOK);
+				}
 				*retval = val;
 				break;
 			}
 		}
-
-		SKIPWS(input, stop);
-		if (input == stop)
-			RETERR(JSON_E_BADDELIM);
-		else if (*input == ',') {
-			++input;
-		} else if (*input == ']' || *input == '}') {
-			continue;
-		} else
-			RETERR(JSON_E_BADDELIM);
 	}
-	*endp = (char*)input;
+
+	if (endp)
+		*endp = (char*)input;
 	return JSON_E_NOERR;
 }
 
+void
+json_context_free(json_context *ctx)
+{
+	struct json_value *v;
+	while ((v = j_context_pop(ctx)) != NULL)
+		json_value_free(v);
+}
+
 int
 json_parse(char const *input, size_t len,
 	   struct json_value **retval, char **endp)
 {
 	struct j_context *ctx = NULL;
-	int rc = json__parse(&ctx, input, input + len, retval, endp);
-	while (j_context_pop(&ctx))
-		;
-	return rc;
-}
+	int rc;
 
+	if (input == NULL || len == 0 || retval == NULL)
+		return JSON_E_USAGE;
+
+	rc = json__parse(&ctx, input, input + len, retval, endp,
+			 NULL);
+	json_context_free(&ctx);
+	return rc & ~JSON_E_MOREINPUT;
+}
 
 int
 json_parse_string(char const *input, struct json_value **retval, char **endp)
 {
 	return json_parse(input, strlen(input), retval, endp);
 }
+
+int
+json_parse_incr(char const *input, size_t len, json_context *ctx,
+		struct json_value **retval, size_t *consumed, char **endp)
+{
+	char *tok, *end;
+	int rc;
+
+	if (retval == NULL || consumed == NULL)
+		return JSON_E_USAGE;
+
+	rc = json__parse(ctx, input, input + len, retval, &end, &tok);
+	if (rc & JSON_E_MOREINPUT)
+		*consumed = tok - input;
+	else
+		*consumed = end - input;
+	if (endp)
+		*endp = end;
+	return rc;
+}
